/* This file contains the variable and function declarations */

# ifndef _GLOBAL_H_
# define _GLOBAL_H_

# define INF 1.0e14
# define EPS 1.0e-14
# define E  2.71828182845905
# define PI 3.14159265358979

/* set the reference point as externer variable */
extern const double reference[];

/* ERROR Process */
#define ERROR(x) fprintf(stderr, x), fprintf(stderr, "\n"), exit(1)

/* ûпԼ⣬ڶṹʱԭʼNSGA-II˼ */

typedef struct
{
    int rank;
    double *xreal;
    double *obj;
    double crowd_dist;
}
individual;

typedef struct
{
    individual *ind;
}
population;

typedef struct lists
{
    int index;
    struct lists *parent;
    struct lists *child;
}
list;

extern int k;
extern int l;
extern int dimension;
extern int nreal;					// ʵĸ
extern int nobj;					// ĿĿ
extern int popsize;				    // ȺС
extern double pcross_real;
extern double pmut_real;
extern double eta_c;
extern double eta_m;
extern nrealmut;					// 
extern int nrealcross;				// 
extern double *min_realvar;			// ʵֵ½
extern double *max_realvar;			// ʵֵϽ
extern double *min_pf;				/* Pareto frontmininmal߽ */
extern double *max_pf;				/* Pareto frontmaximal߽ */

/* 2010/11/1 */
extern double jumping_percentage;	// ȷtransposonı
extern int no_trans;				// transposon
extern double jumping_rate;			// 趨jumpingı
extern int jumping_count;			// jumping

/* 2010/11/15 */
extern int stop_index;				// ʾjumpingȺλ

void allocate_memory (double **front[], int noPoints);
void Deallocate_memrory (double *front[], int noPoints);
void allocate_memory_pop (population *pop, int size);
void allocate_memory_ind (individual *ind);
void deallocate_memory_pop (population *pop, int size);
void deallocate_memory_ind (individual *ind);

double maximum (double a, double b);
double minimum (double a, double b);

void crossover (individual *parent1, individual *parent2, individual *child1, individual *child2);
void realcross (individual *parent1, individual *parent2, individual *child1, individual *child2);

void assign_crowding_distance_list (population *pop, list *lst, int front_size);
void assign_crowding_distance_indices (population *pop, int c1, int c2);
void assign_crowding_distance (population *pop, int *dist, int **obj_array, int front_size);

int check_dominance (individual *a, individual *b);

void evaluate_pop (population *pop, int size);
void evaluate_ind (individual *ind);

void fill_nondominated_sort (population *mixed_pop, population *new_pop);
void crowding_fill (population *mixed_pop, population *new_pop, int count, int front_size, list *cur);

void initialize_pop (population *pop);
void initialize_ind (individual *ind);

void insert (list *node, int x);
list* del (list *node);

void merge(population *pop1, population *pop2, population *pop3);
void copy_ind (individual *ind1, individual *ind2);

void mutation_pop (population *pop);
void mutation_ind (individual *ind);
void real_mutate_ind (individual *ind);

void test_problem (double *xreal, double *obj);

void assign_rank_and_crowding_distance (population *new_pop);

void report_pop (population *pop, FILE *fpt);
void report_feasible (population *pop, FILE *fpt);
void report_ind (individual *ind, FILE *fpt);

void quicksort_front_obj(population *pop, int objcount, int obj_array[], int obj_array_size);
void q_sort_front_obj(population *pop, int objcount, int obj_array[], int left, int right);
void quicksort_dist(population *pop, int *dist, int front_size);
void q_sort_dist(population *pop, int *dist, int left, int right);

void selection (population *old_pop, population *new_pop);
individual* tournament (individual *ind1, individual *ind2);

/* performance metrics */
void get_test_points (double **obj,int num);
double get_gd (population *pop);
double get_igd (population *pop);
double get_hv (population *pop);
void gd_igd_wfg (population *pop, int front_size);
double uniform_diversity (population *pop);

/* Hypervolume */
double CalculateHypervolume (double *front[], int noPoints, int noObjectives);
double CalculateArea (double *front[], int noPoints, int noObjectives);
double CalculateArea_2D (double *front[], int noPoints, int noObjectives);
int Is_In_Set (double *temp[], int temp_length, double *point);
int Dominates (double *point1, double *point2, int noObjectives);
int Check_Dominate (double *front[], int noPoints, int objective);
int FilterNondominatedSet (double *front[], int noPoints, int noObjectives);
double getBound (double *front[], int noPoints, int i, int j);
void Swap (double *front[], int i, int j);
void copy (double *temp[], double *front[], int i, int j);
int IsRepeaded (int x, int array[], int size);
int ReduceFront(double *front[], int noPoints, int noObjectives, double *slice_depth, int slice_number);
void Sort (double *front[], int noPoints, int objective);
static void Merge (double *front[], int p, int q, int r, int objectives);
void Merge_Sort (double *front[], int p, int r, int objectives);

/* 2010/11/1 */
//void JumpingOperation (population *pop, population *jumped_pop, int popsize);
void JumpingOperation (population *pop, population *new_pop, population *jumped_pop, int popsize);
void CutPasteSame (individual *ind);
void CutPasteDiff (individual *ind1, individual *ind2);
void CopyPasteSame (individual *ind);
void CopyPasteDiff (individual *ind1, individual *ind2);
int IsDuplicate (int *trans_position, int check_position, int end);
int IsInTransposon (int *trans_position, int check_position);

/* 2010/11/2 */
int fill_nondominated_sort_2 (population *mixed_pop, population *new_pop, population *jump_pop, population *jumped_pop, population *actual_jumped_pop);
void crowding_fill_2 (population *mixed_pop, population *jump_pop, population *new_pop, int count, int front_size, list *elite, int jump_size);


/* 2010/11/04 */
merge_2 (population *parent_pop, population *jump_pop, population *mixed_pop, int popsize, int jump_size);
void fill_nondominated_sort_3 (population *mixed_pop, population *new_pop, int size);

/* 2011/01/01 */
void score_pop (population *pop_final);
void gd_igd_zdt3 (population *pop);
void gd_igd_dtlz7 (population *pop);

# endif
