#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "global.h"
#include "rand.h"

double get_hv (population *pop)
{
	int i, j;
	int flag;
	int no_feasible;
	
	double **temp;
	double **obj_front;
	double hypervolume;
	
	no_feasible = 0;
	hypervolume = 0;

	/* ڴ */
	allocate_memory (&temp, popsize);

	for (i = 0; i < popsize; i++)
	{
		flag = 0;

		for (j = 0; j < nobj; j++)
		{
			if (pop->ind[i].obj[j] > reference[j])
			{
				flag = 1;
				break;
			}
		}
		if (flag == 0)
		{
			for (j = 0; j < nobj; j++)
			{
				temp[no_feasible][j] = pop->ind[i].obj[j];
				
			}
			no_feasible++;
		}
	}
	printf ("%d\n", no_feasible);
	allocate_memory (&obj_front, no_feasible);

	/* pop */
	for (i = 0; i < no_feasible; i++)
	{
		for (j = 0; j < nobj; j++)
		{
			obj_front[i][j] = temp[i][j];
		}
	}

	if (nobj < 3)
	{
		hypervolume = CalculateArea_2D (obj_front, no_feasible, nobj);
	}
	else
	{
		hypervolume = CalculateHypervolume (obj_front, no_feasible, nobj);
	}

	/* ڴͷ */
	Deallocate_memrory (temp, popsize);
	Deallocate_memrory (obj_front, no_feasible);

	return hypervolume;
}

/* Calculate the hypervolume of 'noPoints' with 'noObjectives' */
double CalculateHypervolume (double *front[], int noPoints, int noObjectives)
{
	int n;
	int slice_number;			/* it is used to record the total number of slices */

	double volume;
	double *slice_depth;		/* this array is used to record the slice depth of the current sliced slices */

	volume = 0;
	n = noPoints;
	slice_number = 0;

	/* this array used to store every depth of slice */
	slice_depth = (double *) malloc (n * n * sizeof(double));
	if (!slice_depth)
		ERROR ("'slilce_depth' memory allocation failed");

	/* the main circle */
	while (n > 0)
	{
		int i;
		int size;
		int temp_length;			
		int noNondominatedPoints;
		int isDominate_flag;	/* this flag is used to denote the domination relation */
		int isInSet_flag;		/* this flag is used to denote whether a point is in a set */

		double tempVolume, area;
		double lowest_upperbound;	
		double **temp;			
		
		tempVolume = 0;
		temp_length = 1;
		size = n;

		/* array temp is used to store the points which are used to calculate the area of this slice */
		allocate_memory (&temp, size);

		if (noObjectives > 3)
			tempVolume = CalculateHypervolume (front, n, noObjectives - 1);
		
		/* sort the remaining points in descending order at the 'noObjectives'th Objective */
		Sort (front, n, noObjectives);

		/* copy the 1st point in front to the array temp */
		copy (temp, front, 0, 0);
		
		/* calculate the current slice */
		lowest_upperbound = getBound (front, noPoints, noObjectives, 0);
		slice_depth[slice_number] = fabs (temp[0][noObjectives - 1] - lowest_upperbound);
		
		if (noObjectives == 3)
		{	/* copy the useful points to the temp array */
			for (i = 1; i < n; i++)
			{	/* pick the points which have the same value at the 3rd dimension */
				if (front[i][noObjectives - 1] == temp[0][noObjectives - 1])
				{
					copy (temp, front, temp_length, i);
					temp_length++;
				}
			}

			/* check whether the point1 is dominated or not at the first 'noObjectives - 1' dimensions */
			isDominate_flag = Check_Dominate (front, n, noObjectives - 1);
			if (isDominate_flag == 1)
			{
				area = CalculateArea (temp, temp_length, noObjectives);
				tempVolume = area;
			}
			else
			{	/* pick the nondominated points at the first 'noObjectives - 1' dimensions */
				noNondominatedPoints = FilterNondominatedSet (front, n, noObjectives - 1);
				for (i = 0; i < noNondominatedPoints; i++)
				{
					/* check whether the nondominated points are already in the temp array or not */
					isInSet_flag = Is_In_Set (temp, temp_length, front[i]);
					if (isInSet_flag == 1)
						continue;
					else
					{
						copy (temp, front, temp_length, i);
						temp_length++;
					}
				}
				area = CalculateArea (temp, temp_length, noObjectives);
				tempVolume = area;
			}
		}

		tempVolume = slice_depth[slice_number] * tempVolume;
		slice_number++;

		/* sum the slice volume */
		volume += tempVolume;
		
		/* cut the useless points */
		n = ReduceFront (front, n, noObjectives, slice_depth, slice_number);
		
		Deallocate_memrory (temp, size);
	}
	
	/* deallocate the memory for the next recurse */
	free (slice_depth);
	return volume;
}

/* calculate the 2D area of the projected points */
double CalculateArea (double *front[], int noPoints, int noObjectives)
{
	int i;
	double diff1, diff2;
	double s_value;
	
	s_value = 0;
	diff1 = 0;
	diff2 = 0;
	
	/* sort the points in descending order at 1st dimension */
	Sort (front, noPoints, 1);

	for (i = 0; i < noPoints; i++)
	{
		if (i == 0)
			diff1 = front[i][noObjectives - 3] - reference[noObjectives - 3];
		else
			diff1 = front[i][noObjectives - 3] - front[i - 1][noObjectives - 3];
		diff2 = front[i][noObjectives - 2] - reference[noObjectives - 2]	;

		s_value += (double)(fabs (diff1) * fabs (diff2));
	}

	return s_value;
}

/* calculate the 2D area of the projected points */
double CalculateArea_2D (double *front[], int noPoints, int noObjectives)
{
	int i;
	double diff1, diff2;
	double s_value;
	
	s_value = 0;
	diff1 = 0;
	diff2 = 0;
	
	/* sort the points in descending order at 1st dimension */
	Sort (front, noPoints, 1);

	for (i = 0; i < noPoints; i++)
	{
		if (i == 0)
			diff1 = front[i][noObjectives - 2] - reference[noObjectives - 2];
		else
			diff1 = front[i][noObjectives - 2] - front[i - 1][noObjectives - 2];
		diff2 = front[i][noObjectives - 1] - reference[noObjectives - 1]	;

		s_value += (double)(fabs (diff1) * fabs (diff2));
	}

	return s_value;
}

/* If this point is already in the array temp, then returns 1, else returns 0 */
int Is_In_Set (double *temp[], int temp_length, double *point)
{
	int i, j;
	int flag1, flag2;

	flag1 = 0;
	flag2 = 0;

	for (i = 0; i < temp_length; i++)
	{
		for (j = 0; j < dimension; j++)
		{
			if (temp[i][j] == point[j])
				flag1 = 1;
			else
			{
				flag2 = 1;
				break;
			}
		}
		if (flag1 == 1 && flag2 == 0)
			break;
		flag1 = 0;
		flag2 = 0;
	}

	if (flag1 == 1 && flag2 == 0)
		return 1;
	else
		return 0;
}

/************************************************************************/
/*  If point1 dominates point2 then return 1							*/
/*  If point2 dominates point1 then return -1							*/
/*  If point1 is non-dominated  to point2 then return 0                  */ 
/************************************************************************/
int Dominates (double *point1, double *point2, int noObjectives)
{
	int  i;
	int  flag1, flag2;
	
	flag1 = 0;
	flag2 = 0;
	
	for (i = 0; i < noObjectives; i++)
	{
	  if (point1[i] <= point2[i])
		  flag1 = 1;
	  else
		  flag2 = 1;
	}
	if (flag1 == 1 && flag2 == 0)
	  return 1;
	else if (flag1 == 0 && flag2 == 1)
	  return -1;
	else
	  return 0;
} /* Dominates */

/* If point 1 is dominated by the other points in the set at the first 'objective' dimension, return 1, else return 0 */
int Check_Dominate (double *front[], int noPoints, int objective)
{
	int i, flag;
	int n;
	
	flag = 0;
	n = noPoints;

	for (i = 1; i < n; i++)
	{
		if (Dominates (front[i], front[0], objective) == 1)
		{
			flag = 1;
			break;
		}
		else
			flag = 0;
	}
	
	return flag;
}

/* filter the front at the first noObjectives - 1 dimension */
int FilterNondominatedSet (double *front[], int noPoints, int noObjectives)
{
	int  i, j;
	int  n;

	n = noPoints;
	i = 0;

	while (i < n) 
	{
		j = i + 1;

		while (j < n) 
		{
			if (Dominates (front[i], front[j], noObjectives)) 
			{
			/* remove point 'j' */
				n--;
				Swap (front, j, n);
			}
			else if (Dominates (front[j], front[i], noObjectives)) 
			{
			/* remove point 'i'; ensure that the point copied to index 'i'
			   is considered in the next outer loop (thus, decrement i) */
				n--;
				Swap (front, i, n);
				i--;
				break;
			}
			else
				j++;
		}
		i++;
	}
	return n;
}

/* Get the j-th point's lowest upper bound at the i-th dimension */
double getBound (double *front[], int noPoints, int i, int j)
{
	int k, bound_length;
	double upperbound, value, threshold;
	double *tempbound;
	
	bound_length = 0;		/* the number of bounds */
	
	/* this array is used to store the bound value of the j-th point */
	tempbound = (double *) malloc ((noPoints + 1) * sizeof(double));
	
	/* threshold is the value of j-th point at i-th dimension */
	threshold = front[j][i - 1];
	
	for (k = 0; k < noPoints; k++)
	{
		value = front[k][i - 1];
		if (value > threshold)
		{
			tempbound[bound_length] = value;
			bound_length++;
		}
	}
	tempbound[bound_length] = reference[i - 1];

	upperbound = tempbound[0];
	for (k = 0; k < bound_length; k++)
	{
		if (tempbound[k] < upperbound)
			upperbound = tempbound[k];	
	}
	
	free (tempbound);
	return upperbound;
}

/* swap the point i and point j */
void Swap (double *front[], int i, int j)
{
	double  *temp;

	temp = front[i];
	front[i] = front[j];
	front[j] = temp;
} /* Swap */

/* copy the j-th point in front to the i-th point in temp */
void copy (double *temp[], double *front[], int i, int j)
{
	int k;

	for (k = 0; k < dimension; k++)
			temp[i][k]=front[j][k];
} /* copy */

/* reduce the useless points in the current front */
int ReduceFront(double *front[], int noPoints, int noObjectives, double *slice_depth, int slice_number)
{
	int n;
	int i;
	double threshold;

	threshold = 0;
	n = noPoints;
	
	for (i = 0; i < slice_number; i++)
		threshold += slice_depth[i];

	for (i = 0; i < n; i++)
	{
		if (fabs (reference[noObjectives - 1] - front[i][noObjectives - 1]) <= threshold)
		{
			n--;
			Swap (front, i, n);
			i--;
		}
	}
	
	return n;
}

/* sort noPoints points in descending order at objective dimension */
void Sort (double *front[], int noPoints, int objective)
{
	Merge_Sort (front, 0, noPoints - 1, objective);
}

static void Merge (double *front[], int p, int q, int r, int objectives)
{
    int i, k;
    int begin1, end1, begin2, end2;
	
	double **temp;
   
	begin1 = p;    
	end1 = q;
    begin2 = q+1;  
	end2 = r;
	
	allocate_memory (&temp, (r - p + 1));

    k = 0;
    while ((begin1 <= end1)&&( begin2 <= end2))
    {
        if (front[begin1][objectives - 1] > front[begin2][objectives - 1])
        {
			copy (temp, front, k, begin1);
			begin1++; 
        }
        else
        {
			copy (temp, front, k, begin2);
            begin2++;
        }
        k++;            
    }
    
    while (begin1 <= end1) 
	{
		copy (temp, front, k, begin1);
		k++;
		begin1++;
	}
	while (begin2 <= end2)
	{
		copy (temp, front, k, begin2);
		k++;
		begin2++;
	}
    
	for (i = 0; i < (r - p +1); i++) 
        copy (front, temp, p + i, i);
	
	Deallocate_memrory (temp, (r - p + 1));
}

void Merge_Sort (double *front[], int p, int r, int objectives)
{
	int q;
	int n;

	n = objectives;

	if (p < r)
	{
		q = (p + r) / 2;

		Merge_Sort (front, p, q, n);
		Merge_Sort (front, q + 1, r, n);
		Merge (front, p, q, r, n);
	}
}