# include <stdio.h>
# include <stdlib.h>
# include <math.h>

# include "global.h"
# include "rand.h"

void matingSelection (population *pop1, int size1, population *pop2, int size2)
{
	selection (pop1, size1, pop2, size2);
	mutation_pop(pop2, size2);
}

void selection (population *old_pop, int size1, population *new_pop, int size2)
{
    int i;
    int rand1, rand2;
    individual *parent1, *parent2;

    for (i = 0; i < size2; i += 2)
    {
		rand1 = rnd (0, size1 - 1);
		rand2 = rnd (0, size1 - 1);
        parent1 = tournament (&old_pop->ind[rand1], &old_pop->ind[rand2]);
		rand1 = rnd (0, size1 - 1);
		rand2 = rnd (0, size1 - 1);
        parent2 = tournament (&old_pop->ind[rand1], &old_pop->ind[rand2]);        
        crossover (parent1, parent2, &new_pop->ind[i], &new_pop->ind[i + 1]);
    }
}

individual* tournament (individual *ind1, individual *ind2)
{
    if (ind1->fitness < ind2->fitness)
    {
        return (ind1);
    }
    if (ind1->fitness > ind2->fitness)
    {
        return (ind2);
    }
    if ((randomperc ()) <= 0.5)
    {
        return (ind1);
    }
    else
    {
        return (ind2);
    }
}

/* Function to cross two individuals */
void crossover (individual *parent1, individual *parent2, individual *child1, individual *child2)
{
	if (nreal != 0)
	{
		realcross (parent1, parent2, child1, child2);
	}
	return;
}

/* Routine for real variable SBX crossover */
void realcross (individual *parent1, individual *parent2, individual *child1, individual *child2)
{
	int i;
	double rand;
	double y1, y2, yl, yu;
	double c1, c2;
	double alpha, beta, betaq;
	if (randomperc() <= pcross_real)
	{
		nrealcross++;
		for (i=0; i<nreal; i++)
		{
			if (randomperc() <= 0.5 )
			{
				if (parent1->xreal[i] < parent2->xreal[i])
				{
					y1 = parent1->xreal[i];
					y2 = parent2->xreal[i];
				}
				else
				{
					y1 = parent2->xreal[i];
					y2 = parent1->xreal[i];
				}
				if (fabs(parent1->xreal[i]-parent2->xreal[i]) > EPS)
				{
					yl = min_realvar[i];
					yu = max_realvar[i];
					rand = randomperc();
					beta = 1.0 + (2.0*(y1-yl)/(y2-y1));
					alpha = 2.0 - pow(beta,-(eta_c+1.0));
					if (rand <= (1.0/alpha))
					{
						betaq = pow ((rand*alpha),(1.0/(eta_c+1.0)));
					}
					else
					{
						betaq = pow ((1.0/(2.0 - rand*alpha)),(1.0/(eta_c+1.0)));
					}
					c1 = 0.5*((y1+y2)-betaq*(y2-y1));
					beta = 1.0 + (2.0*(yu-y2)/(y2-y1));
					alpha = 2.0 - pow(beta,-(eta_c+1.0));
					if (rand <= (1.0/alpha))
					{
						betaq = pow ((rand*alpha),(1.0/(eta_c+1.0)));
					}
					else
					{
						betaq = pow ((1.0/(2.0 - rand*alpha)),(1.0/(eta_c+1.0)));
					}
					c2 = 0.5*((y1+y2)+betaq*(y2-y1));

					if (randomperc()<=0.5)
					{
						child1->xreal[i] = c2;
						child2->xreal[i] = c1;
					}
					else
					{
						child1->xreal[i] = c1;
						child2->xreal[i] = c2;
					}

					if (child1->xreal[i]<yl)
					{
						child1->xreal[i]=yl;
					}
					if (child1->xreal[i]>yu)
					{
						child1->xreal[i]=yu;
					}
					if (child2->xreal[i]<yl)
					{
						child2->xreal[i]=yl;
					}
					if (child2->xreal[i]>yu)
					{
						child2->xreal[i]=yu;
					}
				}
				else
				{
					child1->xreal[i] = parent1->xreal[i];
					child2->xreal[i] = parent2->xreal[i];
				}
			}
			else
			{
				child1->xreal[i] = parent1->xreal[i];
				child2->xreal[i] = parent2->xreal[i];
			}
		}
	}
	else
	{
		for (i=0; i<nreal; i++)
		{
			child1->xreal[i] = parent1->xreal[i];
			child2->xreal[i] = parent2->xreal[i];
		}
	}
	return;
}

void mutation_pop (population *pop,int size)
{
    int i;
    for (i = 0; i < size; i++)
    {
        real_mutate_ind (&(pop->ind[i]));
    }
    return;
}

/* Routine for real polynomial mutation of an individual */
void real_mutate_ind (individual *ind)
{
    int j;
    double rnd, delta1, delta2, mut_pow, deltaq;
    double y, yl, yu, val, xy;
    for (j = 0; j < nreal; j++)
    {
        if (randomperc() <= pmut_real)
        {
            y = ind->xreal[j];
            yl = min_realvar[j];
            yu = max_realvar[j];
            delta1 = (y - yl) / (yu - yl);
            delta2 = (yu - y) / (yu - yl);
            rnd = randomperc ();
            mut_pow = 1.0 / (eta_m + 1.0);
            if (rnd <= 0.5)
            {
                xy = 1.0 - delta1;
                val = 2.0 * rnd + (1.0 - 2.0 * rnd) * (pow (xy, (eta_m + 1.0)));
                deltaq =  pow(val,mut_pow) - 1.0;
            }
            else
            {
                xy = 1.0 - delta2;
                val = 2.0 * (1.0 - rnd) + 2.0 * (rnd - 0.5) * (pow (xy, (eta_m + 1.0)));
                deltaq = 1.0 - (pow(val,mut_pow));
            }
            y = y + deltaq * (yu - yl);
            if (y < yl)
                y = yl;
            if (y > yu)
                y = yu;
            ind->xreal[j] = y;
            nrealmut += 1;
        }
    }
    return;
}